import queue
import threading

import requests

from cloner.put_repos_in_queue import put_repos_in_queue


def obtain_repos(
    github_organization: str,
    github_token: str,
    queue_lock: threading.Lock,
    repo_queue: queue.Queue,
) -> None:
    """
    Makes a request to the GitHub API and then passes that info to the queue.
    """
    github_url = f"https://api.github.com/orgs/{github_organization}/repos"
    headers = {"Accept": "application/vnd.github.v3+json"}

    if github_token is not None:
        headers["Authorization"] = f"token {github_token}"

    # TODO obtain more per page, so less calls are done
    response = requests.get(github_url, headers=headers)
    response.raise_for_status()
    json_response = response.json()

    while "next" in response.links.keys():
        response = requests.get(response.links["next"]["url"], headers=headers)
        json_response.extend(response.json())

    put_repos_in_queue(json_response, queue_lock, repo_queue)



import logging
import queue
import threading

from cloner.repository import Repository

logger = logging.getLogger(__file__)

exit_flag = False
repository_list_queue_lock = None
repository_list_queue = None

# todo - all this class needs tests


def split_queue(
    number_of_threads: int,
    repository_queue: queue.Queue,
    repository_queue_lock: threading.Lock(),
) -> list[list[Repository]]:
    """Splits the queue of repos into a list of repo lists, one per thread"""

    global repository_list_queue_lock, repository_list_queue
    repository_list_queue_lock = repository_queue_lock
    repository_list_queue = repository_queue

    thread_list = []
    repos_to_clone = []

    for i in range(number_of_threads):
        thread = SplitterThread(thread_id=i, total_threads=number_of_threads)
        thread_list.append(thread)
        thread.start()

    while not repository_queue.empty():
        pass

    global exit_flag
    exit_flag = True

    for t in thread_list:
        logger.debug(
            f"Length of thread {t} clone repo list is {len(t.repos_to_clone_list)}"
        )
        t.join()
        repos_to_clone.append(t.repos_to_clone_list)

    return repos_to_clone


class SplitterThread(threading.Thread):
    """Class to split the queue of repos into a list."""

    def __init__(self, thread_id: int, total_threads: int):
        threading.Thread.__init__(self)
        self.thread_id = thread_id
        self.repos_to_clone_list = []
        self.total_threads = total_threads

    def run(self):
        logger.debug(f"Starting thread -> {self.thread_id}")
        self.process_repo()
        logger.debug(
            f"Thread -> {self.thread_id} has {len(self.repos_to_clone_list)} repos"
        )
        logger.debug(f"Exiting thread -> {self.thread_id}")

    def process_repo(self) -> None:
        """
        Obtains a repo from the queue and puts it in its list if the mod of the repo identifier equals this thread id.
        Puts the repo back to the queue otherwise.
        """
        while not exit_flag:
            repository_list_queue_lock.acquire()
            if not repository_list_queue.empty():
                repository = repository_list_queue.get()
                if repository.repo_id % self.total_threads == self.thread_id:
                    self.repos_to_clone_list.append(repository)
                else:
                    repository_list_queue.put(repository)
                repository_list_queue_lock.release()
            else:
                repository_list_queue_lock.release()

    def __str__(self):
        return f"Thread: {self.thread_id}"



import os
from multiprocessing import Process

from cloner.repository import Repository


class ClonerProcess(Process):
    """Class to clone a list of given repositories when started."""

    def __init__(self, repos_to_clone: list[Repository], process_id: int):
        Process.__init__(self)
        self.repos_list = repos_to_clone
        self.process_id = process_id

    def _execute_system_command(self, command):
        return os.system(command=command)

    def run(self):
        """Clones each repo given in the constructor."""
        for repo in self.repos_list:
            # git options? --depth 1  ???
            # todo -> path as an option?
            # https://pypi.org/project/multiprocessing-logging/ not in windows
            self._execute_system_command(
                command=f"git clone {repo.clone_url} --quiet ./repos/{repo.name}"
            )



from cli import cli

if __name__ == "__main__":
    cli()


# same logger for different files
# modules and packages?
# why i cant import this here -> from cloner.cli import cli  ?
# https://docs.python.org/3/tutorial/modules.html#packages



class Repository:
    """Class to abstract the needed attributes of a repository."""

    def __init__(self, name: str, clone_url: str, repo_id: int):
        self.name = name
        self.clone_url = clone_url
        self.repo_id = repo_id

    def __str__(self):
        return f"Repo: {self.name} URL: {self.clone_url}"

    def __eq__(self, other):
        if isinstance(other, Repository):
            return (
                self.name == other.name
                and self.clone_url == other.clone_url
                and self.repo_id == other.repo_id
            )
        return False



import logging

from cloner.cloner_process import ClonerProcess
from cloner.repository import Repository

logger = logging.getLogger(__file__)


def clone_repos(number_of_threads: int, repos_to_clone: list[list[Repository]]) -> None:
    """
    For each given number_of_threads, it creates a ClonerProcess that receives the list in repos_to_clone at that
    the same position.
    If there are more number_of_threads than the length of repos_to_clone, it won't create useless Processes.
    By design, number_of_threads and length of repos_to_clone are equal.

    Note: Cannot be tested (yet) since it needs to get os.system mock per process
    """
    list_of_processes = []
    logger.debug(f"Creating processes to clone repos threads={number_of_threads}")
    for i in range(number_of_threads):
        try:
            process = ClonerProcess(repos_to_clone=repos_to_clone[i], process_id=i)
            list_of_processes.append(process)
            process.start()
        except IndexError:
            pass

    for process in list_of_processes:
        process.join()
    logger.debug("All processes joined")



import queue
import threading
from typing import Any

from cloner.repository import Repository


def put_repos_in_queue(
    json_response: list[dict[str, Any]],
    queue_lock: threading.Lock,
    repo_queue: queue.Queue,
) -> None:
    """
    Puts into the queue repositories obtained from a dictionary
    Each element of the json_response is a dictionary containing the GitHub response with repository information.
    """
    queue_lock.acquire()
    for repo_number in range(len(json_response)):
        repo_queue.put(
            Repository(
                name=json_response[repo_number]["name"],
                clone_url=json_response[repo_number]["clone_url"],
                repo_id=repo_number,
            )
        )
    queue_lock.release()



from pathlib import Path

from setuptools import find_packages, setup

from cloner import __version__

project_root_path = Path(__file__).parent

install_requires = [
    "click>=8.0.3",
    "requests>=2.27.1",
]

dev_requires = [
    "black>=22.3.0",
    "coverage>=6.3.2",
    "faker>=12.0.1",
    "flake8>=4.0.1",
    "isort>=5.10.0",
    "pip-tools>=6.5.1",
    "pytest>=6.2.5",
    "responses>=0.20.0",
]

classifiers = """\
Development Status :: 5 - Production/Stable
Environment :: Console
Intended Audience :: Developers
License :: OSI Approved :: MIT License
Programming Language :: Python
Programming Language :: Python :: 3
Programming Language :: Python :: 3.9
Programming Language :: Python :: 3.10
Programming Language :: Python :: 3.11
"""

# About Python versions. Supported 3.9 and above
# If supporting newer versions, update CI, coverage and classifiers
# If minimum version supported changes, update CI, coverage, classifiers, setup and release.

setup(
    name="wr-cloner",
    version=__version__.__version__,
    author="w0rmr1d3r",
    author_email="",
    description="A tool to clone efficiently all the repos in an organization",
    entry_points={"console_scripts": ["cloner=cloner.cli:cli"]},
    long_description=(project_root_path / "README.md").read_text(),
    long_description_content_type="text/markdown",
    url="https://github.com/w0rmr1d3r/cloner",
    project_urls={},
    classifiers=classifiers.splitlines(),
    packages=find_packages(exclude=("docs", "tests")),
    platforms="any",
    python_requires=">=3.9.0",
    install_requires=install_requires,
    tests_require=dev_requires,
    extras_require={"dev": dev_requires},
)



import logging
import queue
import threading

import click
from requests import HTTPError

from cloner.__version__ import __version__
from cloner.clone_repos import clone_repos
from cloner.obtain_repos import obtain_repos
from cloner.split_queue import split_queue

repository_list_queue_lock = threading.Lock()
repository_list_queue = queue.Queue()

LOGGING_LEVELS = {
    "ERROR": logging.ERROR,
    "WARNING": logging.WARNING,
    "INFO": logging.INFO,
    "DEBUG": logging.DEBUG,
}


def setup_logging(level: str) -> None:
    """Logging setup and configuration."""
    logging.basicConfig(
        level=LOGGING_LEVELS[level], format="%(levelname)s - %(message)s"
    )


@click.command()
@click.version_option(prog_name="cloner", version=__version__)
@click.argument("github_organization")
@click.option(
    "--token",
    "token",
    type=str,
    default=None,
    help="GitHub token to read private repos.",
    show_default=True,
)
@click.option(
    "--threads",
    "threads",
    type=int,
    default=4,
    help="Number of threads and processes to use.",
    show_default=True,
)
@click.option(
    "--logging",
    "logging_level",
    type=click.Choice(LOGGING_LEVELS.keys(), case_sensitive=True),
    default="INFO",
    help="Logging level",
    show_default=True,
)
def cli(github_organization: str, token: str, threads: int, logging_level: str) -> None:
    """Clones all visible repositories for a given organization."""
    setup_logging(level=logging_level)

    logging.info(f"Cloning repos for: {github_organization}")

    try:
        obtain_repos(
            github_organization=github_organization,
            github_token=token,
            queue_lock=repository_list_queue_lock,
            repo_queue=repository_list_queue,
        )
    except HTTPError as e:
        logging.error("An error has occurred while obtaining repos", exc_info=e)

    total_repos_to_clone = repository_list_queue.qsize()
    logging.info(f"Total repos to clone: {total_repos_to_clone}")

    if total_repos_to_clone > 0:
        repos_to_clone = split_queue(
            number_of_threads=threads,
            repository_queue=repository_list_queue,
            repository_queue_lock=repository_list_queue_lock,
        )

        logging.info("Cloning repos...")

        clone_repos(number_of_threads=threads, repos_to_clone=repos_to_clone)

        logging.info("Repos cloned!")


# can this be deleted?
if __name__ == "__main__":
    cli()

# todo -> https://stackoverflow.com/questions/11436502/closing-all-threads-with-a-keyboard-interrupt
